;;;; uuid-ossp.scm
;;;; Kon Lovett, Jan '06

(module uuid-ossp

(;export
  ;
  uuid?
  uuid-null?
  uuid-compare
  uuid=? uuid<? uuid>? uuid<=? uuid>=?
  uuid-copy
  uuid-clear!
  make-uuid
  string->uuid uuid->string
  ;
  uuid-version
  uuid-nil?
  uuid-clone
  uuid-load! uuid-load
  uuid-import uuid-import-binary uuid-import-siv
  uuid-export uuid-export-binary uuid-export-text uuid-export-siv
  ;Deprecated
  uuid<>?)

(import scheme chicken foreign )
(use lolevel)

;;

(define (fxzero? x)
  (fx= 0 x) )

(define (fxnegative? x)
  (fx< x 0) )

(define (fxpositive? x)
  (fx> x 0) )

;;

#>
#include "uuid-ossp-fix.h"
<#

(define UUID_LEN_BIN (foreign-value "uuid_LEN_BIN" unsigned-int))
(define UUID_LEN_STR (foreign-value "uuid_LEN_STR" unsigned-int))
(define UUID_LEN_SIV (foreign-value "uuid_LEN_SIV" unsigned-int))

(define UUID_RC_OK (foreign-value "uuid_RC_OK" unsigned-int))
(define UUID_RC_ARG (foreign-value "uuid_RC_ARG" unsigned-int))
(define UUID_RC_MEM (foreign-value "uuid_RC_MEM" unsigned-int))
(define UUID_RC_SYS (foreign-value "uuid_RC_SYS" unsigned-int))
(define UUID_RC_INT (foreign-value "uuid_RC_INT" unsigned-int))
(define UUID_RC_IMP (foreign-value "uuid_RC_IMP" unsigned-int))

(define UUID_MAKE_V1 (foreign-value "uuid_MAKE_V1" unsigned-int))
(define UUID_MAKE_V1MC (foreign-value "uuid_MAKE_V1MC" unsigned-int))
(define UUID_MAKE_V3 (foreign-value "uuid_MAKE_V3" unsigned-int))
(define UUID_MAKE_V4 (foreign-value "uuid_MAKE_V4" unsigned-int))
(define UUID_MAKE_V5 (foreign-value "uuid_MAKE_V5" unsigned-int))

(define UUID_FMT_BIN (foreign-value "uuid_FMT_BIN" unsigned-int))
(define UUID_FMT_STR (foreign-value "uuid_FMT_STR" unsigned-int))
(define UUID_FMT_TXT (foreign-value "uuid_FMT_TXT" unsigned-int))
(define UUID_FMT_SIV (foreign-value "uuid_FMT_SIV" unsigned-int))

;;

(define-foreign-type size_t "size_t") ;types like this should be collected somewhere

(define-foreign-type uuid_rc_t unsigned-int)
(define-foreign-type uuid_fmt_t unsigned-int)
(define-foreign-type uuid_t (struct "uuid_st"))

(define uuid_create
  (foreign-lambda uuid_rc_t uuid_create (c-pointer (c-pointer uuid_t))))

(define uuid_destroy
  (foreign-lambda uuid_rc_t uuid_destroy (c-pointer uuid_t)))

(define uuid_clone
  (foreign-lambda uuid_rc_t uuid_clone (const (c-pointer uuid_t)) (c-pointer (c-pointer uuid_t))))

(define uuid_load
  (foreign-lambda uuid_rc_t uuid_load (c-pointer uuid_t) (const c-string)))

(define uuid_make_0
  (foreign-lambda uuid_rc_t uuid_make (c-pointer uuid_t) unsigned-int))

(define uuid_make_2
  (foreign-lambda uuid_rc_t uuid_make (c-pointer uuid_t) unsigned-int (c-pointer uuid_t) (const c-string)))

(define uuid_isnil
  (foreign-lambda uuid_rc_t uuid_isnil (const (c-pointer uuid_t)) (c-pointer int)))

(define uuid_compare
  (foreign-lambda uuid_rc_t uuid_compare (const (c-pointer uuid_t)) (const (c-pointer uuid_t)) (c-pointer int)))

(define uuid_import
  (foreign-lambda uuid_rc_t uuid_import (c-pointer uuid_t) uuid_fmt_t (const c-string) size_t))

(define uuid_export
  (foreign-lambda uuid_rc_t uuid_export (const (c-pointer uuid_t)) uuid_fmt_t (c-pointer c-pointer) (c-pointer size_t)))

(define uuid_error
  (foreign-lambda c-string uuid_error uuid_rc_t))

(define uuid_version
  (foreign-lambda unsigned-long uuid_version))

;;

(define (null-pointer? ptr)
  (##sys#check-special ptr 'null-pointer?)
  (zero? (##sys#pointer->address ptr)) )

;;

(define +uuid-error-codes+
  `((,UUID_RC_OK . "everything ok")
    (,UUID_RC_ARG . "invalid argument")
    (,UUID_RC_MEM . "out of memory")
    (,UUID_RC_SYS . "system error")
    (,UUID_RC_INT . "internal error")
    (,UUID_RC_IMP . "not implemented")))

(define (uuid-error-string code)
  (or
    (uuid_error code)
    (let ((msg (assv code +uuid-error-codes+)))
      (if msg
        (cdr msg)
        "unknown result code" ) ) ) )

(define (signal-uuid-error code loc)
  (abort
    (make-composite-condition
      (make-property-condition 'exn 'location loc 'message (uuid-error-string code))
      (make-property-condition 'uuid 'code code))) )

(define-inline (uuid-status-ok? code)
  (fx= UUID_RC_OK code) )

(define-inline (error-check code loc)
  (unless (uuid-status-ok? code)
    (signal-uuid-error code loc) ) )

;;

(define-inline (unbox-puuid boxed-puuid)
  boxed-puuid )

(define (free-uuid uuid)
  (uuid_destroy (unbox-puuid uuid)) )

(define-inline (box-puuid puuid)
  (set-finalizer! (tag-pointer puuid 'uuid-ossp) free-uuid) )

(define (new-uuid loc)
  (let-location (
    (puuid (c-pointer uuid_t)) )
    (error-check (uuid_create (location puuid)) loc)
    (box-puuid puuid) ) )

(define-inline (%uuid? obj)
  (tagged-pointer? obj 'uuid-ossp) )

(define (%uuid-compare uuid1 uuid2 loc)
  (let-location ((comp int))
    (error-check (uuid_compare (unbox-puuid uuid1) (unbox-puuid uuid2) (location comp)) loc)
    comp ) )

;;

(define (uuid-import-format fmt str loc)
  (unless (string? str)
    (##sys#signal-hook
      #:type-error loc
      "bad argument type - not a string" str))
  (let (
    (str-len
      (select fmt
        ((UUID_FMT_BIN) UUID_LEN_BIN)
        ((UUID_FMT_STR) UUID_LEN_STR)
        ((UUID_FMT_SIV) UUID_LEN_SIV)
        (else
          (##sys#signal-hook
            #:type-error loc
            "bad argument type - invalid format" fmt)))) )
    (unless (= (string-length str) str-len)
      ;type-error here is dubious
      (##sys#signal-hook
        #:type-error loc
        "bad argument type - invalid string length" str))
    (let ((uuid (new-uuid loc)))
      (error-check (uuid_import (unbox-puuid uuid) fmt str str-len) loc)
      uuid ) ) )

(define (uuid-export-format uuid fmt loc)
  (let (
    (len-bias
      (select fmt
        ((UUID_FMT_BIN) 0)
        ((UUID_FMT_STR UUID_FMT_TXT UUID_FMT_SIV) 1)
        (else
          (##sys#signal-hook
            #:type-error loc
            "bad argument type - invalid format" fmt)))))
    (let-location (
      (len size_t 0)
      (dat c-pointer #f) )
      (error-check (uuid_export (unbox-puuid uuid) fmt (location dat) (location len)) loc)
      (when (or (null-pointer? dat) (fxzero? len))
        (signal-uuid-error UUID_RC_INT loc))
      (let* (
        (str-len (fx- (inexact->exact len) len-bias))
        (str (make-string str-len)) )
        (move-memory! dat (make-locative str) str-len)
        (free dat)
        str ) ) ) )

(define (get-ns-uuid uuid ns loc)
  (let (
    (ns (or ns "nil")) )
    (unless (string? ns)
      (##sys#signal-hook
        #:type-error loc
        "bad argument type - not a string" ns) )
    (let ((uuid (or uuid (new-uuid loc))))
      (error-check (uuid_load (unbox-puuid uuid) ns) loc)
      uuid ) ) )

(define (make-uuid-0! loc uuid var)
  (error-check (uuid_make_0 (unbox-puuid uuid) var) loc) )

(define (make-uuid-2! loc uuid var args)
  (unless (= 2 (length args))
    (##sys#error-hook
      (foreign-value "C_BAD_MINIMUM_ARGUMENT_COUNT_ERROR" int) loc
      3 (+ 1 (length args)) #f))
  (let* (
    (ns (car args))
    (ns-uuid (if (%uuid? ns) ns (get-ns-uuid #f ns loc)))
    (name (cadr args)) )
    (unless (string? name)
      (##sys#signal-hook
        #:type-error loc
        "bad argument type - not a string" name))
    (error-check (uuid_make_2 (unbox-puuid uuid) var (unbox-puuid ns-uuid) name) loc) ) )

;;

(define-type boxed-uuid pointer)
;(define-type boxed-uuid *)

(: uuid? (* -> boolean : boxed-uuid))
;
(define (uuid? obj)
  (%uuid? obj) )

(: uuid-nil? (* --> boolean))
;
(define (uuid-nil? uuid)
  (and
    (uuid? uuid)
    (let-location (
      (result int) )
      (error-check (uuid_isnil (unbox-puuid uuid) (location result)) 'uuid-nil?)
      (not (zero? result)) ) ) )

(: uuid-null? (* --> boolean))
;
;compatibility w/ uuid-lib
(define uuid-null? uuid-nil?)

(: uuid-compare (boxed-uuid boxed-uuid -> fixnum))
;
(define (uuid-compare uuid1 uuid2)
  (let (
    (cmp (%uuid-compare uuid1 uuid2 'uuid-compare)) )
    (cond
      ((fxnegative? cmp)  -1)
      ((fxzero? cmp)      0)
      (else               1) ) ) )

(: uuid=? (boxed-uuid boxed-uuid --> boolean))
;
(define (uuid=? uuid1 uuid2)
  (fxzero? (%uuid-compare uuid1 uuid2 'uuid=?)) )

(: uuid<>? (boxed-uuid boxed-uuid --> boolean))
;
(define (uuid<>? uuid1 uuid2)
  (not (fxzero? (%uuid-compare uuid1 uuid2 'uuid<>?))) )

(: uuid<? (boxed-uuid boxed-uuid --> boolean))
;
(define (uuid<? uuid1 uuid2)
  (negative? (%uuid-compare uuid1 uuid2 'uuid<?)) )

(: uuid>? (boxed-uuid boxed-uuid --> boolean))
;
(define (uuid>? uuid1 uuid2)
  (positive? (%uuid-compare uuid1 uuid2 'uuid>?)) )

(: uuid<=? (boxed-uuid boxed-uuid --> boolean))
;
(define (uuid<=? uuid1 uuid2)
  (let ((cmp (%uuid-compare uuid1 uuid2 'uuid<=?)))
    (or (fxzero? cmp) (fxnegative? cmp)) ) )

(: uuid>=? (boxed-uuid boxed-uuid --> boolean))
;
(define (uuid>=? uuid1 uuid2)
  (let ((cmp (%uuid-compare uuid1 uuid2 'uuid>=?)))
    (or (fxzero? cmp) (fxpositive? cmp) ) ) )

;;

(: uuid-clone (boxed-uuid --> boxed-uuid))
;
(define (uuid-clone uuid)
  (let-location (
    (puuid (c-pointer uuid_t)) )
    (error-check (uuid_clone (unbox-puuid uuid) (location puuid)) 'uuid-clone)
    (box-puuid puuid) ) )

(: uuid-copy (boxed-uuid --> boxed-uuid))
;
;compatibility w/ uuid-lib
(define uuid-copy uuid-clone)

(: make-uuid (#!optional symbol -> boxed-uuid))
;
(define (make-uuid . args)
  (let (
    (variant (optional args #f))
    (uuid (new-uuid 'make-uuid)) )
    (case variant
      ((#f)
        ) ;the nil-uuid
      ((V1 time)
        (make-uuid-0! 'make-uuid uuid UUID_MAKE_V1))
      ((V1-MC)
        (make-uuid-0! 'make-uuid uuid UUID_MAKE_V1MC))
      ((V3)
        (make-uuid-2! 'make-uuid uuid UUID_MAKE_V3 (cdr args)))
      ((V4 random)
        (make-uuid-0! 'make-uuid uuid UUID_MAKE_V4))
      ((V5)
        (make-uuid-2! 'make-uuid uuid UUID_MAKE_V5 (cdr args)))
      (else
        (##sys#signal-hook
          #:type-error 'make-uuid
          "bad argument type - invalid variant" variant) ) )
    uuid ) )

(: uuid-clear! (boxed-uuid -> boxed-uuid))
;
(define (uuid-clear! uuid)
  (get-ns-uuid uuid "nil" 'uuid-clear!) )

(: uuid-load! (boxed-uuid #!optional string -> boxed-uuid))
;
(define (uuid-load! uuid #!optional ns)
  (get-ns-uuid uuid ns 'uuid-load!) )

(: uuid-load (#!optional string --> boxed-uuid))
;
(define (uuid-load #!optional ns)
  (get-ns-uuid #f ns 'uuid-load) )

(: uuid-import (string --> boxed-uuid))
;
(define (uuid-import str)
  (uuid-import-format UUID_FMT_STR str 'uuid-import) )

(: uuid-import-binary (string --> boxed-uuid))
;
(define (uuid-import-binary str)
  (uuid-import-format UUID_FMT_BIN str 'uuid-import-binary) )

(: uuid-import-siv (string --> boxed-uuid))
;
(define (uuid-import-siv str)
  (uuid-import-format UUID_FMT_SIV str 'uuid-import-binary) )

(: uuid-export (boxed-uuid --> string))
;
(define (uuid-export uuid)
  (uuid-export-format uuid UUID_FMT_STR 'uuid-export) )

(: uuid-export-binary (boxed-uuid --> string))
;
(define (uuid-export-binary uuid)
  (uuid-export-format uuid UUID_FMT_BIN 'uuid-export-binary) )

(: uuid-export-text (boxed-uuid --> string))
;
(define (uuid-export-text uuid)
  (uuid-export-format uuid UUID_FMT_TXT 'uuid-export-text) )

(: uuid-export-siv (boxed-uuid --> string))
;
(define (uuid-export-siv uuid)
  (uuid-export-format uuid UUID_FMT_SIV 'uuid-export-siv) )

(: string->uuid (string --> boxed-uuid))
;
(define (string->uuid str)
  (uuid-import str) )

(: uuid->string (boxed-uuid --> string))
;
(define (uuid->string uuid)
  (uuid-export uuid) )

(: uuid-version (--> number))
;
(define uuid-version uuid_version)

) ;module uuid-ossp
